(** Strategies.
 *
 * We explicit the type for strategies because working with continuations is not particularly readable,
 * hence putting the focus on them and clearly defining their signatures help the understanding.
 *
 * @author Vincent Aravantinos
 *)

type t = 
  |ModelFound of Constraint.t (** since the beginning, at least one model has been found *)
  |NoModelFound               (** no model has been founc until now *)

val unsatisfiable : t -> bool
val satisfiable : t -> bool
val model_hint : t -> Constraint.t option

(*(** Combines two satisfiability datas, intuitive:
  * a model has been found iff a model has been found for any of the two datas
  *)
val (&) : t -> t -> t*)
