open Softcore open OUnit open Schema

let raises_exception ?(exn) f =
  match exn with
  |Some e ->
    (try f (); false
    with e' when e=e' -> true)
  |None   ->
    (try f (); false
    with _ -> true)

let is_regular x = match is_regular x with `Yes -> true | `No _ -> false

module Values =
  struct
    open Schema
    include Constraint_test.Values
    let p                    = Proposition.of_string "P"
    let q                    = Proposition.of_string "Q"
    let int3                 = Indexes.of_num 3
    let p_n                  = pos_lit p termvarn
    let p1                   = pos_lit p int1
    let p0                   = pos_lit p int0
    let q1                   = pos_lit q int1
    let p2                   = pos_lit p int2
    let p3                   = pos_lit p int3
    let p_i                  = pos_lit p termvari
    let p_iplus1             = pos_lit p iplus1
    let q_1                  = pos_lit q int1
    let q_i                  = pos_lit q termvari
    let p_nplus1             = pos_lit p nplus1
    let p_nminus1            = pos_lit p nminus1
    let p_nplus0             = pos_lit p nplus0
    let p_nplus2             = pos_lit p nplus2
    let np0                  = neg_lit p int0
    let np1                  = neg_lit p int1
    let np2                  = neg_lit p int2
    let np_i                 = neg_lit p termvari
    let np_iplus1            = neg_lit p iplus1
    let np_nplus2            = neg_lit p nplus2
    let nq1                  = neg_lit q int1
    let nq_i                 = neg_lit q termvari
    let p1_or_p2             = disj p1 p2
    let p1_or_q1             = disj p1 q1
    let p1_or_pi             = disj p1 p_i
    let pi_or_piplus1        = disj p_i p_iplus1
    let pi_or_npiplus1       = disj p_i np_iplus1
    let npi_or_piplus1       = disj np_i p_iplus1
    let npi_or_npiplus1      = disj np_i np_iplus1
    let p1_or_p2_or_p2       = disj p1_or_p2 p2
    let p1_or_p2_or_p3       = disj p1_or_p2 p3
    let p1_and_p2            = conj p1 p2
    let p1_and_p2_and_p3     = conj p1_and_p2 p3
    let p1_and_q1            = conj p1 q1
    let p1_and_pi            = conj p1 p_i
    let pi_and_piplus1       = conj p_i p_iplus1
    let pi_and_npiplus1      = conj p_i np_iplus1
    let npi_and_piplus1      = conj np_i p_iplus1
    let npi_and_npiplus1     = conj np_i np_iplus1
    let p1_or_p2_and_p2      = conj p1_or_p2 p2
    let itor1                = it_disj vari int1 termvarn  p_i
    let itor1'               = it_disj vari int1 nminus1  p_i
    let itor2                = it_disj vari int1 termvarn  npi_and_piplus1
    let itor3                = it_disj vari int2 nplus1 nq_i
    let itor4                = it_disj vari int0 termvarn npi_or_piplus1
    let itor5                = it_disj vari minus1 termvarn npi_or_piplus1
    let itor6                = it_disj vari int2 nminus2 nq_i
    let itand1               = it_conj vari int1 termvarn p_i
    let itand1'              = it_conj vari int1 nminus1 p_i
    let itand2               = it_conj vari int1 termvarn npi_and_piplus1
    let itand3               = it_conj vari int2 nplus1 nq_i
    let itand4               = it_conj vari int0 termvarn npi_or_piplus1
    let itand5               = it_conj vari minus1 termvarn npi_or_piplus1
    let itand6               = it_conj vari int2 nminus2 nq_i
    let itand7               = it_conj vari int1 termvarn np_i
    let itand8               = it_conj vari int1 termvarn npi_or_piplus1
    let itand9               = it_conj vari int1 nminus1 npi_or_piplus1
    let ititand1             = conj itand1 itand2
    let ititand2             = conj ititand1 itand3
    let ititand3             = conj itand3 ititand1
    let ititand4             = conj ititand3 itand4
    let ititand5             = conj itand6 itand5
    let non_reg1             = it_disj vari int1 termvarn p_n
    let non_reg2             = it_disj vari int1 termvarn (conj p_n p_i)
    let non_reg3             = it_disj vari int1 termvarn (conj p_n p1)
    let two_params1          = conj p_n p_i
    let two_params2          = conj p_n (it_disj vari int1 termvarj p_i)
    let p_vplus1             = pos_lit p vplus1
    let np_vplus1            = neg_lit p vplus1
    let np_nplus1            = neg_lit p nplus1
    let np_nminus1           = neg_lit p nminus1
    let np_n                 = neg_lit p termvarn
    let p_j                  = pos_lit p termvarj
    let pi_or_pj             = disj p_i p_j
    let pnplus1_or_npnplus2  = disj p_nplus1 np_nplus2
    let p1_or_pnplus1        = disj p1 p_nplus1
    let pnplus1_and_npnplus2 = conj p_nplus1 np_nplus2
    let p1_and_pnplus1       = conj p1 p_nplus1
  end

open Values

let test =
  "schema" >:::
    [
      "equal" >::
      (fun () ->
        "p1 pn" @? (not (equal p1 p_n));
        "pn p1" @? (not (equal p_n p1));
      );
      "to_string" >::
      (fun () ->
        "p1"        @? (to_string p_i            = "P_i");
        "p2"        @? (to_string q_i            = "Q_i");
        "p_i"       @? (to_string p1             = "P_1");
        "p_i"       @? (to_string p2             = "P_2");
        "p_iplus1"  @? (to_string p_iplus1       = "P_i+1");
        "np1"       @? (to_string np_i           = "~P_i");
        "np2"       @? (to_string nq_i           = "~Q_i");
        "np_i"      @? (to_string np1            = "~P_1");
        "np_iplus1" @? (to_string np_iplus1      = "~P_i+1");
        "dis1"      @? (to_string p1_or_p2       = "P_1 \\/ P_2");
        "dis2"      @? (to_string p1_or_pi       = "P_1 \\/ P_i");
        "dis3"      @? (to_string pi_or_piplus1  = "P_i \\/ P_i+1");
        "dis4"      @? (to_string pi_or_npiplus1 = "P_i \\/ ~P_i+1");
        "dis5"      @? (to_string npi_or_piplus1 = "~P_i \\/ P_i+1");
      );
    "substitute" >::
      (fun () ->
        "p_i"             @? (substitute ~var:vari ~by:nplus1 p_i             = p_nplus1);
        "np_i"            @? (substitute ~var:vari ~by:nplus1 np_i            = np_nplus1);
        "p1"              @? (substitute ~var:vari ~by:nplus1 p1              = p1);
        "p2"              @? (substitute ~var:vari ~by:nplus1 p2              = p2);
        "p_iplus1t1"      @? (substitute ~var:vari ~by:nplus1 p_iplus1        = p_nplus2);
        "p_iplus1t2"      @? (substitute ~var:vari ~by:nminus1 p_iplus1       = p_n);
        "p_iplus1t3"      @? (substitute ~var:vari ~by:termvarn p_iplus1      = p_nplus1);
        "p_iplus1t4"      @? (substitute ~var:vari ~by:int1 p_iplus1          = p2);
        "ident1"          @? (substitute ~var:varn ~by:nplus1 p_iplus1        = p_iplus1);
        "ident2"          @? (substitute ~var:varn ~by:nplus1 np_iplus1       = np_iplus1);
        "ident3"          @? (substitute ~var:varn ~by:nplus1 p1_or_p2        = p1_or_p2);
        "ident4"          @? (substitute ~var:varn ~by:nplus1 p1_and_p2       = p1_and_p2);
        "in bound 1"      @? (substitute ~var:varn ~by:nminus1 itor1          = itor1');
        "in bound 2"      @? (substitute ~var:varn ~by:nminus1 itand1         = itand1');
        "pi_or_npiplus1"  @? (substitute ~var:vari ~by:nplus1 pi_or_npiplus1  = pnplus1_or_npnplus2);
        "p1_or_pit1"      @? (substitute ~var:vari ~by:nplus1 p1_or_pi        = p1_or_pnplus1);
        "pi_and_npiplus1" @? (substitute ~var:vari ~by:nplus1 pi_and_npiplus1 = pnplus1_and_npnplus2);
        "p1_and_pit1"     @? (substitute ~var:vari ~by:nplus1 p1_and_pi       = p1_and_pnplus1);
        "itor4id"         @? (substitute ~var:vari ~by:nplus1 itor4           = itor4);
        "itand4id"        @? (substitute ~var:vari ~by:nplus1 itand4          = itand4);
        (* substitution in iterations is quite unclear at the moment; should not appear with regular schemata *)
      );
    "max_bound" >::
      (fun () ->
        "no_it"       @? (max_bound p1                         = None);
        "no_it_acc"   @? (max_bound ~acc:(Some nplus1) p1      = Some nplus1);
        "itand1"      @? (max_bound itand1                     = Some termvarn);
        "itor1"       @? (max_bound itor1                      = Some termvarn);
        "itand3"      @? (max_bound itand3                     = Some nplus1);
        "itand6"      @? (max_bound itand6                     = Some nminus2);
        "itand1_acc1" @? (max_bound ~acc:(Some nplus1) itand1  = Some nplus1);
        "itand1_acc2" @? (max_bound ~acc:(Some nminus1) itand1 = Some termvarn);
        "itor1_acc1"  @? (max_bound ~acc:(Some nplus1) itor1   = Some nplus1);
        "itor1_acc2"  @? (max_bound ~acc:(Some nminus1) itor1  = Some termvarn);
        "ititand1"    @? (max_bound ititand1                   = Some termvarn);
        "ititand2"    @? (max_bound ititand2                   = Some nplus1);
        "ititand3"    @? (max_bound ititand3                   = Some nplus1);
        "ititand4"    @? (max_bound ititand4                   = Some nplus1);
        "ititand5"    @? (max_bound ititand5                   = Some termvarn);
      );
    "neg" >::
      (fun () ->
        "p1"              @? (neg p1              = np1);
        "p2"              @? (neg p2              = np2);
        "np1"             @? (neg np1             = p1);
        "np2"             @? (neg np2             = p2);
        "pi_and_piplus1"  @? (neg pi_and_piplus1  = npi_or_npiplus1);
        "npi_or_npiplus1" @? (neg npi_or_npiplus1 = pi_and_piplus1);
        "pi_and_npiplus1" @? (neg pi_and_npiplus1 = npi_or_piplus1);
        "npi_or_piplus1"  @? (neg npi_or_piplus1  = pi_and_npiplus1);
        "itor1"           @? (neg itor1           = itand7);
        "itand7"          @? (neg itand7          = itor1);
      );
    "is_regular" >::
      (fun () ->
        "non_reg1"         @? (not (is_regular non_reg1));
        "non_reg2"         @? (not (is_regular non_reg2));
        "non_reg3"         @? (not (is_regular non_reg3));
        "two_params1"      @? (not (is_regular two_params1));
        "two_params2"      @? (not (is_regular two_params2));
        "p1"               @? (is_regular p1);
        "p_n"              @? (is_regular p_n);
        "p_nplus1"         @? (is_regular p_nplus1);
        "npi_or_piplus1"   @? (is_regular npi_or_piplus1);
        "npi_and_npiplus1" @? (is_regular npi_and_npiplus1);
        "itor1"            @? (is_regular itor1);
        "itand1"           @? (is_regular itand1);
        "ititand1"         @? (is_regular ititand1);
      );
    "looping set" >:::
      let empty_set = LoopingMap.empty in
      let add set sch = LoopingMap.add set ~key:sch ~data:() in
      let p1_set = add        empty_set p1 in
      let np_nplus1_set = add empty_set np_nplus1 in
      let np_n_set = add      empty_set np_n in
      let itand8_set = add    empty_set itand8 in
      let some_set1 = add (add p1_set np_nplus1) itand8 in
      [
        "loops_among" >::
          (fun () ->
            "P_1 -> P_1"                                   @? (Option.is_some (LoopingMap.find p1_set p1));
            "~P_n -> ~P_n+1"                               @? (not (Option.is_some (LoopingMap.find np_nplus1_set np_n)));
            "~P_n -> ~P_n"                                 @? (Option.is_some (LoopingMap.find np_n_set np_n));
            "/\\i=1..n-1 P_i -> P_i+1 -> /\\i=1..n ..."    @? (not (Option.is_some (LoopingMap.find itand8_set itand9)));
            "/\\i=1..n P_i -> P_i+1 -> /\\i=1..n ..."      @? (Option.is_some (LoopingMap.find itand8_set itand8));
            "P_1 -> {P_1;...}"                             @? (Option.is_some (LoopingMap.find some_set1 p1));
            "~P_n+1 -> {~P_n+1;...}"                       @? (Option.is_some (LoopingMap.find some_set1 np_nplus1));
            "/\\i=1..n-1 P_i -> P_i+1 -> {/\\i=1..n-1;..}" @? (Option.is_some (LoopingMap.find some_set1 itand8));
          );
      ]
    ]

